/*
 * @(#) $Id:  $
 *
 * Copyright 2009/2010 by sIT Solutions,
 * A-1110 Wien, Geiselbergstr.21-25.
 * All rights reserved.
 *
 */
package org.openxma.dsl.platform.exceptions;

import java.util.Set;

import javax.validation.ConstraintViolation;

import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import org.springframework.validation.FieldError;
import org.springframework.validation.ObjectError;

/**
 * Supports creating exceptions from validation errors. 
 */
public class ExceptionSupport {

    public static BeanValidationException mapToValidationException(Errors errors) {
        return mapToValidationException(errors, null);
    }

    /**
     * Maps Spring validation errors to an XMA BeanValidationException
     * which contains global validation errors as BeanValidationExceptions
     * and field validation errors as PropertyValidationException.
     * 
     */
    public static BeanValidationException mapToValidationException(Errors errors, Object target) {
        if (target == null && errors instanceof BeanPropertyBindingResult) {
            target = ((BeanPropertyBindingResult) errors).getTarget();
        }
        String beanName = null;
        if (target != null) {
            beanName = target.getClass().getCanonicalName();
        } else {
            beanName = errors.getObjectName();
        }
        BeanValidationException bvex = new BeanValidationException(beanName,
                "Invalid " + beanName + System.getProperty("line.separator") + errors);
        if (errors.hasGlobalErrors()) {
            for (ObjectError objectError : errors.getGlobalErrors()) {
                bvex.addValidationIssue(mapToValidationException(objectError, beanName));
            }
        }
        if (errors.hasFieldErrors()) {
            for (FieldError fieldError : errors.getFieldErrors()) {
                bvex.addValidationIssue(mapToValidationException(fieldError, beanName));
            }
        }
        return bvex;
    }

    /**
     * Maps a Spring object validation error to an XMA {@link BeanValidationException}
     */
    public static BeanValidationException mapToValidationException(ObjectError error, String beanFullName) {
        BeanValidationException bvex = new BeanValidationException(beanFullName, error.getDefaultMessage());
        return bvex;
    }

    /**
     * Maps a Spring field validation error to an XMA {@link PropertyValidationException}
     */
    public static PropertyValidationException mapToValidationException(FieldError error, String beanFullName) {
        PropertyValidationException pvex = null;
        Object[] arguments = error.getArguments();
        if (arguments != null && arguments.length > 0) {
            pvex = new PropertyValidationException(error.getCode(), beanFullName, error.getField(),
                    error.getRejectedValue(), arguments, error.toString());
        } else {
            pvex = new PropertyValidationException(error.getCode(), beanFullName, error.getField(),
                    error.getRejectedValue(), error.toString());
        }
        return pvex;
    }
    

    /**
     * Maps JSR 303 constraint violations to an XMA BeanValidationException
     * which contains global validation errors as BeanValidationExceptions
     * and field validation errors as PropertyValidationException.
     * 
     */
    
    public static BeanValidationException mapToValidationException(Set<? extends ConstraintViolation<?>> violations, Object target) {
        if (violations == null || violations.isEmpty()) {
            return null;
        }
        String beanName = null;
        if (target != null) {
            beanName = target.getClass().getCanonicalName();
        }
        BeanValidationException bvex = null;

        for (ConstraintViolation<?> violation : violations) {
            beanName = violation.getLeafBean().getClass().getCanonicalName();
            
            if (bvex == null) {
                // take first bean name
                bvex = new BeanValidationException(beanName, "Invalid " + beanName + System.getProperty("line.separator") + violations);
            }
            
            String messageTemplate = violation.getMessageTemplate().trim();
            if (messageTemplate.length() > 2 && messageTemplate.charAt(0) == '{' && messageTemplate.indexOf('}') == messageTemplate.length() - 1) {
                messageTemplate = messageTemplate.substring(1, messageTemplate.length() - 1);
            }
            String fieldName = violation.getPropertyPath().toString();
            BeanValidationException valException;
            
            if (fieldName == null || fieldName.length() == 0) {
                valException = new BeanValidationException(messageTemplate, beanName, violation.getMessage());
            } else {
                String defaultMessage = "'" + fieldName + "' "  + violation.getMessage();
                Object value = violation.getInvalidValue();
                valException = new PropertyValidationException(messageTemplate, beanName, fieldName, value, defaultMessage);
            }
            bvex.addValidationIssue(valException);
        }
        return bvex;
    }


}
