/*******************************************************************************
 * Copyright (c) 2010 s IT Solutions AT Spardat GmbH . All rights reserved. This program and the accompanying materials
 * are made available under the terms of the Eclipse Public License v1.0 which accompanies this distribution, and is
 * available at http://www.eclipse.org/legal/epl-v10.html
 * 
 * Contributors: s IT Solutions AT Spardat GmbH - initial API and implementation
 *******************************************************************************/
package org.openxma.dsl.platform.hibernate;

import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.LinkedHashSet;
import java.util.Map;
import java.util.Set;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.hibernate.HibernateException;
import org.hibernate.Query;
import org.hibernate.QueryException;
import org.hibernate.Session;
import org.hibernate.SessionFactory;
import org.hibernate.engine.NamedQueryDefinition;
import org.hibernate.engine.SessionFactoryImplementor;
import org.openxma.dsl.platform.dao.impl.GenericDaoHibernateImpl;
import org.openxma.dsl.platform.hibernate.hql.HqlRewriteStrategy;
import org.openxma.dsl.platform.hibernate.hql.impl.RewritingHqlParser;
import org.openxma.dsl.platform.query.Expression;
import org.openxma.dsl.platform.query.ExpressionOperator;
import org.openxma.dsl.platform.query.QueryContext;
import org.openxma.dsl.platform.query.QueryObject;
import org.openxma.dsl.platform.query.impl.BinaryExpression;
import org.openxma.dsl.platform.query.impl.ConstantExpression;
import org.openxma.dsl.platform.query.impl.LogicalExpression;
import org.openxma.dsl.platform.query.impl.ParameterExpression;
import org.openxma.dsl.platform.query.impl.RelationalExpression;
import org.openxma.dsl.platform.query.impl.SortOrderExpression;
import org.openxma.dsl.platform.service.Mapper;

import antlr.ANTLRException;

public class NamedQueryFilter implements QueryContext, HqlRewriteStrategy {
    /** Logger available to subclasses */
    protected final static Log logger = LogFactory.getLog(NamedQueryFilter.class);

    private static final Map<ExpressionOperator, String> HQL_OPERATOR_MAP = new HashMap<ExpressionOperator, String>() {
        private static final long serialVersionUID = 1L;

        {
            put(ExpressionOperator.AND, "and");
            put(ExpressionOperator.ASC, "asc");
            put(ExpressionOperator.DESC, "desc");
            put(ExpressionOperator.EQUALS, "=");
            put(ExpressionOperator.IS_NULL, "is");
            put(ExpressionOperator.IS_NOT_NULL, "is not");
            put(ExpressionOperator.GREATER_THAN, ">");
            put(ExpressionOperator.GREATER_THAN_OR_EQUAL, ">=");
            put(ExpressionOperator.LESS_THAN, "<");
            put(ExpressionOperator.LESS_THAN_OR_EQUAL, "<=");
            put(ExpressionOperator.LIKE, "like");
            put(ExpressionOperator.NOT_EQUALS, "!=");
            put(ExpressionOperator.NOT_LIKE, "not like");
            put(ExpressionOperator.OR, "or");
        }

    };
    private StringBuilder filterExpressionBuilder = new StringBuilder();
    private StringBuilder sortOrderExpressionBuilder = new StringBuilder();
    private Map<String, Object> parameterMap = new HashMap<String, Object>();
    private int parameterIndex;
    private String currentVariableName;
    private RewritingHqlParser rewritingHqlParser;
    private SessionFactoryImplementor sessionFactoryImplementor;
    private Mapper namedQueryMapper;

    public NamedQueryFilter(SessionFactory sessionFactory, Mapper namedQueryMapper) {
        super();
        this.sessionFactoryImplementor = (SessionFactoryImplementor) sessionFactory;
        this.namedQueryMapper = namedQueryMapper;
    }

    public Map<String, Object> getParameters() {
        return parameterMap;
    }

    public Object accept(QueryObject queryObject) {
        copyArguments(queryObject);
        buildFilterExpression(queryObject);
        buildSortOrderHql(queryObject);
        return list(queryObject, rewriteHqlQuery(queryObject));
    }

    protected void buildFilterExpression(QueryObject queryObject) {
        if (null != queryObject.getWhereExpression()) {
            queryObject.getWhereExpression().evaluate(this);
        }
    }

    protected Collection list(QueryObject queryObject, String rewrittenHqlString) throws HibernateException {
        if (logger.isDebugEnabled()) {
            logger.debug(rewrittenHqlString);
        }
        Query query = createQuery(queryObject, rewrittenHqlString);
        query.setReadOnly(true);
        if (rewritingHqlParser.isPropertiesSelect()) {
            return namedQueryMapper.createAndMapMany(query.list(), null, Collections.singletonMap(
                    GenericDaoHibernateImpl.QUERY_NAME, queryObject.getQueryName()));
        }
        return rewritingHqlParser.getJoinAliasMap().isEmpty() ? query.list() : new LinkedHashSet(query.list());
    }

    protected Query createQuery(QueryObject queryObject, String rewrittenHqlString) throws HibernateException {
        Session currentSession = this.sessionFactoryImplementor.getCurrentSession();
        Query query = currentSession.createQuery(rewrittenHqlString);
        applyParameters(query);
        applyQueryHints(queryObject, query);
        return query;
    }

    protected void applyQueryHints(QueryObject queryObject, Query query) {
        query.setFirstResult(queryObject.getFirstResult());
        if (queryObject.getMaxResults() > 0) {
            query.setMaxResults(queryObject.getMaxResults());
        }
    }

    private void applyParameters(Query query) throws HibernateException {
        Set<String> parameterNames = getParameters().keySet();
        for (String parameterName : parameterNames) {
            Object val = getParameters().get(parameterName);
            if (val instanceof Collection) {
                query.setParameterList(parameterName, (Collection) val);
            } else {
                query.setParameter(parameterName, val);                
            }
        }
    }

    private void buildSortOrderHql(QueryObject queryObject) {
        for (Expression expression : queryObject.getSortOrderExpressions()) {
            expression.evaluate(this);
        }
    }

    private String rewriteHqlQuery(QueryObject queryObject) {
        NamedQueryDefinition namedQueryDefinition = sessionFactoryImplementor.getNamedQuery(queryObject.getQueryName());
        if (namedQueryDefinition == null) {
            throw new QueryException("NameQuery '" + queryObject.getQueryName() + "' not found");
        }
        String hqlQueryString = namedQueryDefinition.getQueryString();
        rewritingHqlParser = RewritingHqlParser.getInstance(hqlQueryString, sessionFactoryImplementor);
        rewritingHqlParser.setRewritingStrategy(this);
        try {
            rewritingHqlParser.statement();
        } catch (ANTLRException e) {
            throw new QueryException(e.getMessage(), hqlQueryString);
        }
        return rewritingHqlParser.rewriteHqlQuery();
    }

    protected void copyArguments(QueryObject queryObject) {
        Set<String> parameterNames = queryObject.getArguments().keySet();
        for (String parameterName : parameterNames) {
            getParameters().put(parameterName, queryObject.getArguments().get(parameterName));
        }
    }

    public Object accept(RelationalExpression relationalExpression) {
        Object name = relationalExpression.getLeft().evaluate(this);
        currentVariableName = name + String.valueOf(parameterIndex++);
        appendOperator(relationalExpression);
        Object value = relationalExpression.getRight().evaluate(this);
        parameterMap.put(toParameterName(currentVariableName), value);
        return null;
    }

    public Object accept(LogicalExpression logicalExpression) {
        if (logicalExpression.getParent() != null) {
            filterExpressionBuilder.append("(");
        }
        logicalExpression.getLeft().evaluate(this);
        appendOperator(logicalExpression);
        logicalExpression.getRight().evaluate(this);
        if (logicalExpression.getParent() != null) {
            filterExpressionBuilder.append(")");
        }
        return null;
    }

    public Object accept(ConstantExpression constantExpression) {
        if (ConstantExpression.NULL.equals(constantExpression)) {
            filterExpressionBuilder.append(ConstantExpression.NULL.toString());
        } else {
            filterExpressionBuilder.append(":").append(toParameterName(currentVariableName));
        }
        return constantExpression.getValue();
    }

    public Object accept(ParameterExpression parameterExpression) {
        filterExpressionBuilder.append(prefixWithAliasPlaceHolder(parameterExpression.getName()));
        return parameterExpression.getName();
    }

    public Object accept(SortOrderExpression sortOrderExpression) {
        ParameterExpression parameterExpression = (ParameterExpression) sortOrderExpression.getExpression();
        String fieldName = parameterExpression.getName();
        if (sortOrderExpressionBuilder.length() > 0) {
            sortOrderExpressionBuilder.append(",");
        }
        sortOrderExpressionBuilder.append(prefixWithOrderByAliasPlaceHolder(fieldName)).append(" ").append(
                HQL_OPERATOR_MAP.get(sortOrderExpression.getOperator()));
        return null;
    }

    public String getFilter() {
        return filterExpressionBuilder.toString();
    }

    public String getSortOrder() {
        return sortOrderExpressionBuilder.toString();
    }

    protected String toParameterName(String path) {
        return path.replaceAll("\\.", "_");
    }

    private Object prefixWithOrderByAliasPlaceHolder(String fieldName) {
        return prefixWithAliasPlaceHolder(RewritingHqlParser.ORDER_BY_ALIAS_PLACEHOLDER, fieldName);
    }
    
    private Object prefixWithAliasPlaceHolder(String fieldName) {
        return prefixWithAliasPlaceHolder(RewritingHqlParser.ALIAS_PLACEHOLDER, fieldName);
    }
    
    private Object prefixWithAliasPlaceHolder(String alias, String fieldName) {
        return String.format("%s.%s", alias, fieldName);
    }

    private void appendOperator(BinaryExpression binaryExpression) {
        filterExpressionBuilder.append(" ");
        filterExpressionBuilder.append(HQL_OPERATOR_MAP.get(binaryExpression.getOperator()));
        filterExpressionBuilder.append(" ");
    }

}
