package org.openxma.dsl.platform.service.impl;

import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import org.openxma.dsl.platform.assembler.DtoAssembler;
import org.openxma.dsl.platform.assembler.DtoAssemblerRegistry;
import org.openxma.dsl.platform.dao.EntityFactoryRegistry;
import org.openxma.dsl.platform.service.Mapper;
import org.openxma.dsl.platform.service.MapperContextHolder;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.stereotype.Component;

/**
 * Adapter implementation to adapt the {@link Mapper} strategy interface to classes implementing {@link DtoAssembler}.
 * THis class
 */
@Component
@SuppressWarnings("unchecked")
public class EntitySupportMapper implements Mapper {
    @Autowired
    private EntityFactoryRegistry entityFactoryRegistry;

    @Autowired
    private DtoAssemblerRegistry dtoAssemblerRegistry;

    public <T> List<T> createAndMapMany(Collection<?> sourceCollection, Class<T> targetClass) {
        return createAndMapMany(sourceCollection, targetClass, new HashMap<Object, Object>());
    }

    public <T> T mapOne(Object sourceObject, T targetObject) {
        return mapOne(sourceObject, targetObject, new HashMap<Object, Object>());
    }

    public <T> T createAndMapOne(Object sourceObject, Class<T> targetClass) {
        return createAndMapOne(sourceObject, targetClass, new HashMap<Object, Object>());
    }

    public <T> T createAndMapOne(Object sourceObject, final Class<T> targetClass, final Object context) {
        if (null == targetClass)
            throw new IllegalArgumentException("parameter 'targetClass' must not be null");

        if (null == sourceObject)
            return null;

        final Object newSourceObject = extractSourceObjectFromCollection(sourceObject);
        if (null == newSourceObject)
            return null;

        return executeCreateAndMapOne(newSourceObject, targetClass, context);
    }

    private <T> T executeCreateAndMapOne(final Object sourceObject, final Class<T> targetClass, final Object context) {
        return execute(new MappingCommand<T>() {
            public T execute() {
                return createAndMapOneInternal(sourceObject, targetClass);
            }

            public Object getContext() {
                return context;
            }
        });
    }

    private Object extractSourceObjectFromCollection(Object sourceObject) {
        if (!(sourceObject instanceof Collection))
            return sourceObject;

        Collection<Object> sourceCollection = (Collection<Object>) sourceObject;
        if (sourceCollection.isEmpty())
            return null;
        else
            return sourceCollection.iterator().next();
    }

    private <T> T createAndMapOneInternal(Object sourceObject, Class<T> targetClass) {
        DtoAssembler<T> targetDtoAssembler = dtoAssemblerRegistry.getDtoAssembler(targetClass);
        DtoAssembler<T> sourceDtoAssembler = dtoAssemblerRegistry.getDtoAssembler((Class<T>) sourceObject.getClass());

        if (MapperContextHolder.isMapperOperation() || null != sourceDtoAssembler)
            return sourceDtoAssembler.mapTo((T) sourceObject, targetClass);
        else if (null != targetDtoAssembler)
            return targetDtoAssembler.mapFromEntity(sourceObject);
        else
            throw new IllegalArgumentException("missing DtoAssembler");
    }

    public <T> List<T> createAndMapMany(Collection<?> sourceCollection, Class<T> targetClass, Object context) {
        if (null == targetClass) {
            throw new IllegalArgumentException("parameter 'targetClass' must not be null");
        }

        List<T> result = new ArrayList<T>();
        if (null == sourceCollection)
            return result;

        for (Object object : sourceCollection) {
            T mapOne = createAndMapOne(object, targetClass, context);
            if (mapOne != null)
                result.add(mapOne);
        }
        return result;
    }

    public <T> T mapOne(final Object sourceObject, final T targetObject, final Object context) {
        if (null == targetObject)
            throw new IllegalArgumentException("parameter 'targetObject' must not be null");

        if (null == sourceObject)
            return targetObject;

        return executeMapOne(sourceObject, targetObject, context);
    }

    private <T> T executeMapOne(final Object sourceObject, final T targetObject, final Object context) {
        return execute(new MappingCommand<T>() {
            public T execute() {
                return mapOneInternal(sourceObject, targetObject);
            }

            public Object getContext() {
                return context;
            }
        });
    }

    private <T> T mapOneInternal(Object sourceObject, T targetObject) {
        DtoAssembler<T> targetDtoAssembler = dtoAssemblerRegistry.getDtoAssembler((Class<T>) targetObject.getClass());
        DtoAssembler<T> sourceDtoAssembler = dtoAssemblerRegistry.getDtoAssembler((Class<T>) sourceObject.getClass());

        if (MapperContextHolder.isMapperOperation() || null != sourceDtoAssembler)
            sourceDtoAssembler.mapTo((T) sourceObject, (Object) targetObject);
        else if (null != targetDtoAssembler) {
            targetObject = (T) targetDtoAssembler.mapFromEntity(sourceObject, targetObject);
        } else
            throw new IllegalArgumentException("missing DtoAssembler");

        return targetObject;
    }

    private <T> T execute(MappingCommand<T> mappingCommand) {
        MapperContextHolder.setRegistryWithContext(entityFactoryRegistry, getContextMap(mappingCommand.getContext()),
                false);
        try {
            return mappingCommand.execute();
        } finally {
            MapperContextHolder.reset();
        }
    }

    private Map<Object, Object> getContextMap(Object context) {
        Map<Object, Object> result = new HashMap<Object, Object>();
        if (context instanceof Map) {
            result = (Map<Object, Object>) context;
        } else if (null != context) {
            result.put(MapperContextHolder.CONTEXT, context);
        }
        return result;
    }
}
