/*
 * @(#) $Id:  $
 *
 * Copyright 2009/2010 by sIT Solutions,
 * A-1110 Wien, Geiselbergstr.21-25.
 * All rights reserved.
 *
 */
package org.openxma.dsl.platform.validation;

import java.text.MessageFormat;
import java.util.Iterator;
import java.util.Locale;
import java.util.NoSuchElementException;

import org.springframework.context.support.ResourceBundleMessageSource;

/**
 * MessageSource for resolving validation error messages generated for DataView and Entity constraints.
 * <br/><br/>
 * The resource bundles to search have to be specified as resource bundle names (see {@link #setBasenames(String[])}). Additionally, 
 * the following default openXMA bundles are added at the end of this list for default values:
 * <br/>
 * {@value #DEFAULT_VALIDATION_MESSAGES}</li>
 * <br/>
 * {@value #XMA_VALIDATION_MESSAGES}</li>
 * <br/><br/>
 * Message codes are resolved in the following way:
 * <ul>
 *   <li>The given code is resolved and, if successful, the resolved message is returned</li>
 *   <li>If no message was found, the code is stripped off the last segment (separated by '.') and tried again</li>
 *   <li>If no code segment is available anymore, <code>null</code> is returned</li>
 * </ul>
 * <br/>
 * As an example, consider the message code <code>validation.Size.String.book.subTitle</code>. The code resolution will be in
 * the following order, until a match is found:
 * <ul>
 *   <li><code>validation.Size.String.book.subTitle</code></li>
 *   <li><code>validation.Size.String.book</code></li>
 *   <li><code>validation.Size.String</code></li>
 *   <li><code>validation.Size</code></li>
 *   <li><code>validation</code></li>
 * </ul>
 * 
 * @since 6.0
 */
public class Jsr303ValidationMessageSource extends ResourceBundleMessageSource {

    private static final String DEFAULT_VALIDATION_MESSAGES = "org.openxma.dsl.platform.validation.DefaultValidationMessages";
    private static final String XMA_VALIDATION_MESSAGES = "org.openxma.dsl.platform.validation.XMAValidationMessages";

    /* (non-Javadoc)
     * @see org.springframework.context.support.ResourceBundleMessageSource#setBasenames(java.lang.String[])
     * 
     * Add default XMA validation message bundle at the end of the list
     */
    @Override
    public void setBasenames(String[] basenames) {
        String[] defaultBasenames = new String[basenames.length + 2];
        System.arraycopy(basenames, 0, defaultBasenames, 0, basenames.length);
        defaultBasenames[defaultBasenames.length - 2] = DEFAULT_VALIDATION_MESSAGES;
        defaultBasenames[defaultBasenames.length - 1] = XMA_VALIDATION_MESSAGES;
        super.setBasenames(defaultBasenames);
    }

    @Override
    protected String resolveCodeWithoutArguments(String code, Locale locale) {
        for (Itr iterator = new Itr(code); iterator.hasNext(); ) {
            String thisCode = iterator.next();
            String resolved = super.resolveCodeWithoutArguments(thisCode, locale);
            if (resolved != null) {
                return resolved;
            }
        }
        return null;
    }
    
    @Override
    protected MessageFormat resolveCode(String code, Locale locale) {
        for (Itr iterator = new Itr(code); iterator.hasNext(); ) {
            String thisCode = iterator.next();
            MessageFormat resolved = super.resolveCode(thisCode, locale);
            if (resolved != null) {
                return resolved;
            }
        }
        return null;
    }
    
    private class Itr implements Iterator<String> {
        
        private final String message;
        private int currentIndex;

        public Itr(String message) {
            this.message = message;
            if (message != null) {
                currentIndex = message.length();
            }
        }

        public boolean hasNext() {
            if (message == null) {
                return false;
            }
            return currentIndex > 0;
        }

        public String next() {
            if (message == null) {
                throw new NoSuchElementException();
            }
            String result = message.substring(0, currentIndex);
            currentIndex = result.lastIndexOf('.');
            return result;
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
    
}
