package org.openxma.dsl.platform.validation;

import java.lang.annotation.Annotation;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.validation.ConstraintValidator;
import javax.validation.ConstraintValidatorContext;
import javax.validation.ConstraintValidatorContext.ConstraintViolationBuilder;
import javax.validation.ConstraintViolation;
import javax.validation.Path;
import javax.validation.Path.Node;
import javax.validation.Validation;
import javax.validation.Validator;
import javax.validation.groups.Default;

import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;
import org.springframework.beans.factory.annotation.Autowired;

/**
 * Base class for JSR 303 validators.
 * 
 * If the validators are managed by spring, the {@link javax.validation.Validator} has to be configured in your
 * application context for autowiring. Otherwise a fallback validator is created using default configuration settings.
 *
 * @param <V>
 * @param <E>
 */
public abstract class Jsr303Validators<V extends Annotation, E> extends BaseValidators implements ConstraintValidator<V, E> {

    protected final Log logger = LogFactory.getLog(getClass());
    
    private static Validator fallbackValidator;
    
    @Autowired
    private Validator validator;
    
    public void setValidator(Validator validator) {
        this.validator = validator;
    }
    
    public void initialize(V constraintAnnotation) {
    }

    public boolean isValid(E object, ConstraintValidatorContext context) {
        Set<ConstraintViolation<E>> violations = validateConditional(object);
        if (violations.isEmpty()) {
            return true;
        }
        context.disableDefaultConstraintViolation();
        for (ConstraintViolation<E> constraintViolation : violations) {
            
            Path propertyPath = constraintViolation.getPropertyPath();
            Iterator<Node> pathIterator = propertyPath.iterator();
            if (pathIterator.hasNext()) {
                Node node = pathIterator.next();
                String name = node.getName();
                addFieldViolation(name, constraintViolation.getMessageTemplate(), context);
                // what to do with more nodes? also add them?
            } else {
                addViolation(constraintViolation.getMessageTemplate(), context);
            }
        }
        return false;
    }
    
    protected void addViolation(String messageTemplate, ConstraintValidatorContext context) {
        addFieldViolation(null, messageTemplate, context);
    }
    
    protected void addFieldViolation(String fieldName, String messageTemplate, ConstraintValidatorContext context) {
        ConstraintViolationBuilder violationBuilder = context.buildConstraintViolationWithTemplate(messageTemplate);
        if (fieldName == null) {
            violationBuilder.addConstraintViolation();
        } else {
            violationBuilder.addNode(fieldName).addConstraintViolation();
        }
    }
    
    /**
     * Validate all constraints of the given object, including {@link Default} group and conditional groups.
     * @param object object to validate
     * @return constraint violations or empty Set
     */
    public Set<ConstraintViolation<E>> validate(E object) {
        return getValidator().validate(object);
    }
    
    /**
     * Validate conditional groups constraints of the given object, excluding {@link Default} group constraints.
     * @param object object to validate
     * @return Set of constraint violations or empty Set
     */
    protected Set<ConstraintViolation<E>> validateConditional(E object) {
        Set<Class<?>> groups = createGroups(object);
        groups.remove(Default.class);
        if (groups.isEmpty()) {
            return new HashSet<ConstraintViolation<E>>();
        }
        return getValidator().validate(object, groups.toArray(new Class<?>[groups.size()]));
    }
    
    /**
     * Evaluate which validation groups should be included in the conditional validation process.
     * @param object object which is to be validated
     * @return Set of groups to validate for given object. The {@link Default} group should not be included.
     */
    protected Set<Class<?>> createGroups(E object) {
        return new HashSet<Class<?>>();
    }
    
    private Validator getValidator() {
        if (validator == null) {
            synchronized (Jsr303Validators.class) {
                // TODO try to get spring managed validator?
                if (fallbackValidator == null) {
                    if (logger.isErrorEnabled()) {
                        logger.error("Validator not managed by Spring, generating fallback validator for: " + this);
                    }
                    fallbackValidator = Validation.byDefaultProvider().configure().buildValidatorFactory().getValidator();
                }
                validator = fallbackValidator;
            }
        }
        return validator;
    }
    

}
