package org.openxma.dsl.platform.validation;

import java.math.BigDecimal;
import java.util.Locale;
import java.util.Set;

import javax.validation.ConstraintViolation;
import javax.validation.constraints.DecimalMax;
import javax.validation.constraints.DecimalMin;
import javax.validation.constraints.Digits;
import javax.validation.constraints.NotNull;
import javax.validation.constraints.Size;

import org.springframework.context.i18n.LocaleContextHolder;
import org.springframework.util.Assert;
import org.springframework.validation.BeanPropertyBindingResult;
import org.springframework.validation.Errors;
import org.springframework.validation.MessageCodesResolver;
import org.springframework.validation.ValidationUtils;

import at.spardat.enterprise.exc.SysException;
import at.spardat.enterprise.fmt.FmtFactory;
import at.spardat.enterprise.fmt.FmtParseException;
import at.spardat.enterprise.fmt.IFmt;

public final class ValidatorsHelper {
    
    private ValidatorsHelper() {
    }
    
    /**
     * Reject the given string value with the given error code and error
     * arguments if the length is greater than the given maximum.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the field value to check
     * @param maxLength
     *            the maximum (inclusive) field length
     */
    public static void rejectIfMaxLength(Errors errors, String name, String value, int maxLength) {
        if (value != null && value.trim().length() > maxLength) {
            errors.rejectValue(name, "field.max.length", new Object[] { name, value, maxLength }, "Attribute '" + name
                    + "' must have at most [" + maxLength + "] characters.");
        }
    }

    /**
     * Reject the given string value with the given error code and error
     * arguments if the length is lower than the given minimum length.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the field value to check
     * @param minLength
     *            the minimum (inclusive) field length
     */
    public static void rejectIfMinLength(Errors errors, String name, String value, int minLength) {
        if (value != null && value.trim().length() < minLength) {
            errors.rejectValue(name, "field.min.length", new Object[] { name, value, minLength }, "Attribute '" + name
                    + "' must have at least [" + minLength + "] characters.");
        }
    }

    /**
     * Reject the given field value with the given error code and error
     * arguments if the value is greater than the given max value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the number value to check
     * @param maxValue
     *            the maximum (inclusive) field length
     */
    public static void rejectIfMaxValue(Errors errors, String name, Number value, int maxValue) {
        if (value != null && value.intValue() > maxValue) {
            errors.rejectValue(name, "field.max.value", new Object[] { name, value, maxValue }, "Attribute '" + name
                    + "' must not be bigger than [" + maxValue + "].");
        }
    }

    /**
     * Reject the given number value with the given error code and error
     * arguments if the value is lower than the given min value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the number value to check
     * @param minValue
     *            the maximum (inclusive) field length
     */
    public static void rejectIfMinValue(Errors errors, String name, Number value, int minValue) {
        if (value != null && value.intValue() < minValue) {
            errors.rejectValue(name, "field.min.value", new Object[] { name, value, minValue }, "Attribute '" + name
                    + "' must not be smaller than [" + minValue + "].");
        }
    }

    /**
     * Reject the given number value with the given error instance if the number
     * of integral digits is greater than the given max value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the number value to check
     * @param maxIntegerDigits
     *            the maximum (inclusive) integral digits for the number
     */
    public static void rejectIfMaxIntegerDigits(Errors errors, String name, Number value, int maxIntegerDigits) {
        BigDecimal bigDecimal = toBigDecimal(value);
        int integerPartLength = bigDecimal.precision() - bigDecimal.scale();
        if (integerPartLength > maxIntegerDigits) {
            errors.rejectValue(name, "field.max.integer.digits", new Object[] { name, value, maxIntegerDigits },
                    "Attribute '" + name + "' must have at most [" + maxIntegerDigits
                            + "] digits in the integer portion.");
        }
    }

    /**
     * Reject the given number value with the given error instance if the number
     * of fractional digits is greater than the given max value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the number value to check
     * @param maxFractionalDigits
     *            the maximum (inclusive) fractional digits for the number
     */
    public static void rejectIfMaxFractionDigits(Errors errors, String name, Number value, int maxFractionalDigits) {
        BigDecimal bigDecimal = toBigDecimal(value);
        int fractionPartLength = bigDecimal.scale() < 0 ? 0 : bigDecimal.scale();
        if (fractionPartLength > maxFractionalDigits) {
            errors.rejectValue(name, "field.max.fractional.digits", new Object[] { name, value, maxFractionalDigits },
                    "Attribute '" + name + "' must have at most [" + maxFractionalDigits
                            + "] digits in the fraction portion.");
        }
    }

    /**
     * Reject the given number value with the given error code and error
     * arguments if the value is lower than the given min value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the value to check
     * @param format
     *            the format to apply
     */
    public static void rejectIfInvalidFormat(Errors errors, String name, String value, String format) {
        rejectIfInvalidFormat(errors, name, value, FmtFactory.create(format, getLocale()));
    }

    /**
     * Reject the given number value with the given error code and error
     * arguments if the value is lower than the given min value.
     * 
     * @param errors
     *            the Errors instance to register errors on
     * @param name
     *            the field name to check
     * @param value
     *            the value to check
     * @param format
     *            the format to apply
     */
    public static void rejectIfInvalidFormat(Errors errors, String name, String value, IFmt format) {
        try {
            format.parse(value);
        } catch (FmtParseException fmtParseException) {
            errors.rejectValue(name, getErrorCode(fmtParseException), fmtParseException.getUIMessage(getLocale()));
        }
    }
    
    /**
     * @return the default <code>Locale</code>
     */
    protected static Locale getLocale() {
        return LocaleContextHolder.getLocale();
    }

    /**
     * @param value
     *            the number to convert
     * @return a new BigDecimal instance for the given number value
     */
    protected static BigDecimal toBigDecimal(Number value) {
        BigDecimal result = null;
        if (value instanceof BigDecimal) {
            result = (BigDecimal) value;
        } else if (value != null) {
            result = new BigDecimal(value.toString());
        }
        return null != result ? result.stripTrailingZeros() : BigDecimal.ZERO;
    }

    // TODO replace this
    private static String getErrorCode(FmtParseException fmtParseException) {
        try {
            return (String) fmtParseException.getClass().getField("bundleKey_").get(fmtParseException);
        } catch (Exception e) {
            throw new SysException(e);
        }
    }
    
    /**
     * @param target
     *            target the target bean to validate
     * @return a new errors container
     */
    public static Errors createErrors(Object target, MessageCodesResolver messageCodesResolver) {
        Assert.notNull(target, "parameter 'target' must not be null");
        BeanPropertyBindingResult beanPropertyBindingResult = new BeanPropertyBindingResult(target, target.getClass()
                .getSimpleName());
        beanPropertyBindingResult.setMessageCodesResolver(messageCodesResolver);
        return beanPropertyBindingResult;
    }
    

    /**
     * Convert JSR 303 constraint violations to Spring Errors
     * @param violations Constraint violations as received from JSR 303 validation
     * @param object Object which was validated
     * @param messageCodesResolver message code resolver for Spring error messages
     * @return
     */
    public static Errors toErrors(Set<? extends ConstraintViolation<?>> violations, Object object, MessageCodesResolver messageCodesResolver) {
        Errors errors = createErrors(object, messageCodesResolver);
        
        for (ConstraintViolation<?> constraintViolation : violations) {
            toErrors(errors, constraintViolation, object);
        }
        return errors;
    }
    
    public static void toErrors(Errors errors, ConstraintViolation<?> violation, Object object) {
        String name = violation.getPropertyPath().toString();
        Object annotation = violation.getConstraintDescriptor().getAnnotation();
        Object value = violation.getInvalidValue();
        
        if (annotation instanceof org.hibernate.validator.constraints.NotEmpty) {
            ValidationUtils.rejectIfEmptyOrWhitespace(errors, name, "field.required");
            
        } else if (annotation instanceof NotNull) {
            ValidationUtils.rejectIfEmptyOrWhitespace(errors, name, "field.required");
            
        } else if (annotation instanceof DecimalMin) {
            if (value instanceof Number) {
                String minValue = ((DecimalMin) annotation).value();
                rejectIfMinValue(errors, name, (Number) value, new BigDecimal(minValue).intValue());
            }
            
        } else if (annotation instanceof DecimalMax) {
            if (value instanceof Number) {
                String maxValue = ((DecimalMax) annotation).value();
                rejectIfMaxValue(errors, name, (Number) value, new BigDecimal(maxValue).intValue());
            }
            
        } else if (annotation instanceof Digits) {
            if (value instanceof Number) {
                Digits digitsAnnotation = (Digits) annotation;
                int maxIntegerDigits = digitsAnnotation.integer();
                int maxFractionalDigits = digitsAnnotation.fraction();
                
                rejectIfMaxIntegerDigits(errors, name, (Number) value, maxIntegerDigits);
                rejectIfMaxFractionDigits(errors, name, (Number) value, maxFractionalDigits);
            }
            
        } else if (annotation instanceof Size) {
            if (value instanceof String) {
                Size sizeAnnotation = (Size) annotation;
                rejectIfMinLength(errors, name, (String) value, sizeAnnotation.min());
                rejectIfMaxLength(errors, name, (String) value, sizeAnnotation.max());
            }
        }
    }
}
