package org.openxma.dsl.platform.xma;

import javax.servlet.http.HttpSession;

import org.openxma.dsl.platform.exceptions.ApplicationException;
import org.openxma.dsl.platform.exceptions.SystemException;
import org.openxma.dsl.platform.i18n.MessageProvider;
import org.openxma.dsl.platform.i18n.SimpleErrorMessageProvider;
import org.springframework.beans.BeansException;
import org.springframework.beans.factory.NoSuchBeanDefinitionException;
import org.springframework.beans.factory.config.AutowireCapableBeanFactory;
import org.springframework.context.ApplicationContext;
import org.springframework.web.context.WebApplicationContext;
import org.springframework.web.context.support.WebApplicationContextUtils;

import at.spardat.enterprise.exc.AppException;
import at.spardat.enterprise.exc.SysException;
import at.spardat.xma.component.ComponentServer;
import at.spardat.xma.page.PageServer;
import at.spardat.xma.session.XMASessionServer;

/**
 * An abstract {@link ComponentServer} class which overrides the default implementation of
 * method {@link ComponentServer#newPageModel(short)} to delegate the actual creation of the
 * <code>PageServer</code> instances to spring i.e. returns an instance, which may be shared
 * or independent, of the specified bean from a configured spring context.
 *
 * @see org.springframework.web.context.support.WebApplicationContextUtils
 * @see org.springframework.web.context.ContextLoaderListener
 */
public abstract class SpringComponentServer extends DslComponentServer {
	protected ApplicationContext applicationContext;
	protected MessageProvider messageProvider = new SimpleErrorMessageProvider();
	
	public SpringComponentServer(XMASessionServer session, boolean isStateless,
			short id) {
		super(session, isStateless, id);
		this.applicationContext = getWebApplicationContext();
		this.applicationContext.getAutowireCapableBeanFactory()
			.autowireBeanProperties(this, AutowireCapableBeanFactory.AUTOWIRE_AUTODETECT, true);
	}

	/**
	 * @see at.spardat.xma.component.ComponentServer#newPageModel(short)
	 */
	@Override
    public PageServer newPageModel (short typeId) {
    	PageServer newPageModel = super.newPageModel(typeId);
    	this.applicationContext.getAutowireCapableBeanFactory()
			.autowireBeanProperties(newPageModel, AutowireCapableBeanFactory.AUTOWIRE_AUTODETECT, true);
    	return newPageModel;
    }

	/**
	 * Return an instance, which may be shared or independent, of the specified bean.
	 * @return an instance of the bean
	 * @throws NoSuchBeanDefinitionException if there is no bean definition
	 * with the specified name
	 * @throws BeansException if the bean could not be obtained
	 */
	public Object getBean(String name) {
		return getWebApplicationContext().getBean(name);
	}

	/**
	 * @return the root WebApplicationContext for this web app
	 */
	public WebApplicationContext getWebApplicationContext() {
		HttpSession httpSession = getSession().getHttpSession();
    	WebApplicationContext webApplicationContext = WebApplicationContextUtils
				.getRequiredWebApplicationContext(httpSession.getServletContext());
		return webApplicationContext;
	}

	/**
	 * Converting platform base exceptions to Sys-/AppExceptions while resolving error codes to messages.
	 */
	@Override
    public Throwable convertToBaseException(Throwable detail) {
        if (detail instanceof ApplicationException) {
            return new AppException(detail, this.messageProvider.getMessage((ApplicationException) detail,
                    getSession().getContext().getLocale()));
        } else if (detail instanceof SystemException) {
            return new SysException(detail, this.messageProvider.getMessage((SystemException) detail,
                    getSession().getContext().getLocale()));
        }
        return super.convertToBaseException(detail);
    }
	
	public void setMessageProvider(MessageProvider messageProvider) {
	    this.messageProvider = messageProvider;
	}
		
}
